# copyright 2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-seda unit tests for schema"""

from json import dumps
import unittest

from six import text_type

from cubicweb.devtools.testlib import CubicWebTC
from cubicweb.web import INTERNAL_FIELD_VALUE

from cubicweb_seda.xsd2yams import RULE_TYPES
from cubicweb_seda.views import archiveunit, clone, export, mgmt_rules

from cubicweb_seda import testutils


def entity_fields(fields):
    return [text_type(field.name) for field in fields
            if field.eidparam and not field.name.startswith('__')]


class ManagementRulesTC(CubicWebTC):
    def test_rule_ref_vocabulary(self):
        with self.admin_access.cnx() as cnx:
            create = cnx.create_entity

            access_scheme = create('ConceptScheme', title=u'access')
            access_concept = access_scheme.add_concept(label=u'anyone')
            cnx.commit()

            bdo = testutils.create_transfer_to_bdo(cnx)
            transfer = bdo.container[0]

            rule_base = create('SEDAAccessRule', seda_access_rule=transfer)
            rule_alt = create('SEDAAltAccessRulePreventInheritance',
                              reverse_seda_alt_access_rule_prevent_inheritance=rule_base)
            cnx.create_entity('SEDAPreventInheritance', seda_prevent_inheritance=rule_alt)
            cnx.commit()

            self.assertEqual(mgmt_rules._rule_ref_vocabulary(transfer, rule_base.cw_etype),
                             [('you must specify a scheme for seda_access_rule_code_list_version_'
                               'from_object to select a value', INTERNAL_FIELD_VALUE)])
            self.assertEqual(mgmt_rules._rule_ref_vocabulary(transfer, rule_alt.cw_etype),
                             [('you must specify a scheme for seda_access_rule_code_list_version_'
                               'from_object to select a value', INTERNAL_FIELD_VALUE)])

            create('SEDAAccessRuleCodeListVersion',
                   seda_access_rule_code_list_version_from=transfer,
                   seda_access_rule_code_list_version_to=access_scheme)
            cnx.commit()

            self.assertEqual(mgmt_rules._rule_ref_vocabulary(transfer, rule_base.cw_etype),
                             [('<no value specified>', '__cubicweb_internal_field__'),
                              (access_concept.label(), text_type(access_concept.eid))])
            self.assertEqual(mgmt_rules._rule_ref_vocabulary(transfer, rule_alt.cw_etype),
                             [('<no value specified>', '__cubicweb_internal_field__'),
                              (access_concept.label(), text_type(access_concept.eid))])

    def test_archive_unit_component_rule_ref_vocabulary(self):
        with self.admin_access.cnx() as cnx:
            unit, unit_alt, unit_alt_seq = testutils.create_archive_unit(None, cnx=cnx)
            for rule_type in ('access', 'appraisal'):
                etype = 'SEDASeq{0}RuleRule'.format(rule_type.capitalize())
                scheme = testutils.scheme_for_type(cnx, 'seda_rule', etype)
                concept = scheme.add_concept(label=u'whatever')
                cnx.commit()

                unit, unit_alt, unit_alt_seq = testutils.create_archive_unit(None, cnx=cnx)

                self.assertEqual(mgmt_rules._rule_ref_vocabulary(unit_alt_seq, etype),
                                 [('<no value specified>', '__cubicweb_internal_field__'),
                                  ('whatever', text_type(concept.eid))])


class PermissionsTC(CubicWebTC):
    """Functional test case about permissions in the web interface."""

    def test_container_permissions(self):
        """Check that a user cannot edit a SEDA profile he/she did not create."""
        alice_login = 'alice'
        bob_login = 'bob'
        with self.admin_access.cnx() as cnx:
            self.create_user(cnx, login=alice_login)
            self.create_user(cnx, login=bob_login)
            cnx.commit()
        with self.new_access(alice_login).web_request() as req:
            alice_profile = req.create_entity('SEDAArchiveTransfer', title=u'Alice Profile')
            req.cnx.commit()
        with self.new_access(bob_login).web_request() as req:
            bob_profile = req.create_entity('SEDAArchiveTransfer', title=u'Bob Profile')
            req.cnx.commit()
        with self.new_access(alice_login).web_request() as req:
            alice_profile = req.entity_from_eid(alice_profile.eid)
            bob_profile = req.entity_from_eid(bob_profile.eid)
            alice_profile_form = self.vreg['forms'].select('edition', req, entity=alice_profile)
            bob_profile_form = self.vreg['forms'].select('edition', req, entity=bob_profile)
            alice_field_names = [field.name for field in alice_profile_form.fields]
            bob_field_names = [field.name for field in bob_profile_form.fields]
            for field_name in ('title',
                               'user_annotation',
                               'seda_archival_agency',
                               'seda_transferring_agency'):
                self.assertIn(field_name, alice_field_names)
                self.assertNotIn(field_name, bob_field_names)


class RelationWidgetTC(CubicWebTC):
    """Functional test case about the relation widget."""

    def test_linkable_rset(self):
        with self.admin_access.cnx() as cnx:
            transfer = cnx.create_entity('SEDAArchiveTransfer', title=u'Test widget')
            bdo = cnx.create_entity('SEDABinaryDataObject',
                                    user_annotation=u'I am mandatory',
                                    seda_binary_data_object=transfer)
            bdo_alt = cnx.create_entity('SEDAAltBinaryDataObjectAttachment',
                                        reverse_seda_alt_binary_data_object_attachment=bdo)
            cnx.create_entity('SEDAAttachment', seda_attachment=bdo_alt)
            compressed = cnx.create_entity('SEDACompressed', seda_compressed=bdo)
            cnx.commit()
        with self.admin_access.web_request() as req:
            compressed = req.entity_from_eid(compressed.eid)
            req.form = {'relation': 'seda_algorithm:Concept:subject',
                        'container': text_type(transfer.eid)}
            view = self.vreg['views'].select('search_related_entities', req,
                                             rset=compressed.as_rset())
            self.failIf(view.linkable_rset())
        with self.admin_access.cnx() as cnx:
            scheme = cnx.create_entity('ConceptScheme', title=u'CompressionAlgorithm')
            scheme.add_concept(label=u'bz2')
            cnx.create_entity('SEDACompressionAlgorithmCodeListVersion',
                              seda_compression_algorithm_code_list_version_from=transfer,
                              seda_compression_algorithm_code_list_version_to=scheme)
            cnx.commit()
        with self.admin_access.web_request() as req:
            compressed = req.entity_from_eid(compressed.eid)
            req.form = {'relation': 'seda_algorithm:Concept:subject',
                        'container': text_type(transfer.eid)}
            view = self.vreg['views'].select('search_related_entities', req,
                                             rset=compressed.as_rset())
            self.assertEqual(len(view.linkable_rset()), 1)


class HelperFunctionsTC(CubicWebTC):
    def test_rtags_from_xsd_element(self):
        from cubicweb.web.views.uicfg import reledit_ctrl
        from cubicweb_seda.views import rtags_from_xsd_element

        rsection, display_ctrl = rtags_from_xsd_element('SEDABinaryDataObject', 'FileInfo')

        self.assertEqual(rsection.etype_get('SEDABinaryDataObject', 'filename', 'subject'),
                         None)
        self.assertEqual(reledit_ctrl.etype_get('SEDABinaryDataObject', 'filename', 'subject'),
                         {'novalue_include_rtype': False, 'novalue_label': u'<no value specified>'})


class FormatSupportedPredicateTC(CubicWebTC):

    def test_format_supported(self):
        class fakecls(object):
            def __init__(self, seda_version):
                self.seda_version = seda_version

        predicate = export.format_supported()

        with self.admin_access.web_request() as req:
            transfer = req.create_entity('SEDAArchiveTransfer', title=u'Test widget',
                                         simplified_profile=True)
            req.cnx.commit()

            for seda_version in ('2.0', '1.0', '0.2'):
                self.assertEqual(predicate(fakecls(seda_version), req, entity=transfer),
                                 1)

            self.assertEqual(predicate(fakecls(seda_version='1000'), req, entity=transfer),
                             0)

            with req.cnx.security_enabled(write=False):
                transfer.cw_set(compat_list=u'SEDA 2.0')
                req.cnx.commit()
            transfer.cw_clear_all_caches()

            self.assertEqual(predicate(fakecls(seda_version='2.0'), req, entity=transfer),
                             1)

            req.form['version'] = '2.0'  # not considered
            self.assertEqual(predicate(fakecls(seda_version='1.0'), req, entity=transfer),
                             0)

            req.form['version'] = '2.0'
            self.assertEqual(predicate(None, req, entity=transfer),
                             1)

            req.form['version'] = '1.0'
            self.assertEqual(predicate(None, req, entity=transfer),
                             0)

            del req.form['version']
            self.assertEqual(predicate(fakecls(seda_version='1.0'), req, entity=transfer,
                                       version='2.0'),
                             0)
            self.assertEqual(predicate(None, req, entity=transfer, version='2.0'),
                             1)
            self.assertEqual(predicate(None, req, entity=transfer, version='1.0'),
                             0)

            transfer.cw_set(simplified_profile=False)
            req.cnx.commit()  # will reset compat_list

            self.assertEqual(predicate(None, req, entity=transfer, version='2.0'),
                             1)
            self.assertEqual(predicate(None, req, entity=transfer, version='1.0'),
                             0)


class ArchiveTransferDiagnoseTabTC(CubicWebTC):

    def test_diagnose_tab(self):
        with self.admin_access.web_request() as req:
            transfer = req.create_entity('SEDAArchiveTransfer', title=u'diagnosis testing')
            unit, unit_alt, unit_alt_seq = testutils.create_archive_unit(transfer)
            req.cnx.commit()
            # ensure the diagnosis tab display correctly
            self.view('seda_at_diagnose_tab', req=req, rset=transfer.as_rset())


class ArchiveTransferExportTC(CubicWebTC):

    def test_selected_export_class(self):
        with self.admin_access.web_request() as req:
            transfer = req.create_entity('SEDAArchiveTransfer', title=u'diagnosis testing',
                                         simplified_profile=True)
            unit, unit_alt, unit_alt_seq = testutils.create_archive_unit(transfer)
            req.cnx.commit()

            exporter = self.vreg['views'].select('seda.export', req, entity=transfer)
            self.assertEqual(exporter.__class__, export.SEDAExportView)

            req.form['version'] = '0.2'
            exporter = self.vreg['views'].select('seda.export', req, entity=transfer)
            self.assertEqual(exporter.__class__, export.SEDAExportView)

            req.form['version'] = '3.0'
            exporter = self.vreg['views'].select('seda.export', req, entity=transfer)
            self.assertEqual(exporter.__class__, export.SEDANonSupportedExportView)

            with req.cnx.security_enabled(write=False):
                transfer.cw_set(compat_list=u'SEDA 2.0')
                req.cnx.commit()

            req.form['version'] = '0.2'
            exporter = self.vreg['views'].select('seda.export', req, entity=transfer)
            self.assertEqual(exporter.__class__, export.SEDANonSupportedExportView)

            req.form['version'] = '2.0'
            exporter = self.vreg['views'].select('seda.export', req, entity=transfer)
            self.assertEqual(exporter.__class__, export.SEDAExportView)

    def test_export_filename(self):
        with self.admin_access.web_request() as req:
            transfer = req.create_entity('SEDAArchiveTransfer',
                                         title=u'diagnosis testing',
                                         simplified_profile=True)
            req.cnx.commit()
            for version, fmt, expected_filename in (
                ('2.0', 'rng', 'diagnosis testing-2.0.rng'),
                ('2.0', 'html', 'diagnosis testing-2.0.html'),
                ('1.0', 'rng', 'diagnosis testing-1.0.rng'),
                ('1.0', 'xsd', 'diagnosis testing-1.0.xsd'),
            ):
                req.form['version'], req.form['format'] = version, fmt
                view = self.vreg['views'].select('seda.export', req,
                                                 rset=transfer.as_rset())
                view.set_request_content_type()
                self.assertEqual(
                    view._cw.headers_out.getRawHeaders('content-disposition'),
                    ['attachment;filename="{0}"'.format(expected_filename)])


class SimplifiedFormsTC(CubicWebTC):
    """Functional test case about forms in the web interface."""

    def setup_database(self):
        with self.admin_access.cnx() as cnx:
            transfer = cnx.create_entity('SEDAArchiveTransfer', title=u'diagnosis testing',
                                         simplified_profile=True)
            archive_unit = testutils.create_archive_unit(transfer)[0]

            cnx.commit()
            self.transfer_eid = transfer.eid
            self.archive_unit_eid = archive_unit.eid

    def create_and_link_form(self, req, etype, linkto=None, **kwargs):
        # add minimal information in form to mimick __linkto behaviour
        if linkto is None:
            linkto = self.archive_unit_eid
        req.form['__linkto'] = 'x:{0}:y'.format(linkto)
        return self._create_etype_form(req, etype, **kwargs)

    def create_inlined_form(self, req, etype, parent=None, **kwargs):
        # add minimal information in form to mimick inlined form behaviour
        if parent is None:
            parent = self.archive_unit_eid
        req.form['arg'] = [dumps(text_type(parent))]
        return self._create_etype_form(req, etype, **kwargs)

    def _create_etype_form(self, req, etype, **kwargs):
        entity = req.vreg['etypes'].etype_class(etype)(req)
        return self.vreg['forms'].select('edition', req, entity=entity, **kwargs)

    def assertInlinedFields(self, form, expected):
        """Assert the given inlined form as the expected set of inlined forms, each defined by its
        field name and associated view class'name.
        """
        # text_type around field.name because it may be a relation schema instead of a string
        inlined_fields = [(text_type(field.name), field.view.__class__.__name__)
                          for field in form.fields if hasattr(field, 'view')]
        self.assertEqual(set(inlined_fields), set(expected))

    def assertNoRemove(self, form, field_name, role):
        """Assert the given inlined form field will be rendered without a link to remove it."""
        field = form.field_by_name(field_name, role=role)
        self.assertEqual(field.view._get_removejs(), None)

    def assertNoTitle(self, form, field_name, role):
        """Assert the given inlined form field will be rendered without form title."""
        field = form.field_by_name(field_name, role=role)
        self.assertEqual(field.view.form_renderer_id, 'notitle')

    def test_top_level_rule_form(self):
        with self.admin_access.web_request() as req:
            req.entity_from_eid(self.transfer_eid).cw_set(simplified_profile=False)
            req.cnx.commit()
            for rule_type in RULE_TYPES:
                if rule_type == 'classification':
                    continue
                with self.subTest(rule_type=rule_type):
                    form = self.create_and_link_form(
                        req, 'SEDA{0}Rule'.format(rule_type.capitalize()),
                        linkto=self.transfer_eid)
                    self.assertInlinedFields(form, [
                        ('seda_seq_{0}_rule_rule'.format(rule_type),
                         'InlineAddNewLinkView'),
                    ])
            form = self.create_and_link_form(
                req, 'SEDAClassificationRule'.format(rule_type.capitalize()),
                linkto=self.transfer_eid)
            self.assertInlinedFields(form, [
                ('seda_seq_classification_rule_rule', 'InlineAddNewLinkView'),
                ('seda_classification_reassessing_date', 'InlineAddNewLinkView'),
                ('seda_need_reassessing_authorization', 'InlineAddNewLinkView'),
            ])

    def test_top_level_rule_form_simplified(self):
        with self.admin_access.web_request() as req:
            for rule_type in ('appraisal', 'access'):
                with self.subTest(rule_type=rule_type):
                    form = self.create_and_link_form(
                        req, 'SEDA{0}Rule'.format(rule_type.capitalize()),
                        linkto=self.transfer_eid)
                    self.assertInlinedFields(form, [
                        ('seda_seq_{0}_rule_rule'.format(rule_type),
                         'RuleRuleInlineEntityCreationFormView'),
                    ])

    def test_rule_form(self):
        with self.admin_access.web_request() as req:
            for rule_type in ('access', 'appraisal'):
                form = self.create_and_link_form(req, 'SEDA{0}Rule'.format(rule_type.capitalize()))
                self.assertInlinedFields(form, [
                    ('seda_seq_{0}_rule_rule'.format(rule_type),
                     'RuleRuleInlineEntityCreationFormView'),
                ])
                self.assertNoRemove(form, 'seda_seq_{0}_rule_rule'.format(rule_type), 'subject')

    def test_rule_rule_form(self):
        with self.admin_access.web_request() as req:
            for rule_type in ('access', 'appraisal'):
                form = self.create_and_link_form(
                    req, 'SEDASeq{0}RuleRule'.format(rule_type.capitalize()))
                other_fields = [text_type(field.name)
                                for field in form.fields if not hasattr(field, 'view')]
                self.assertNotIn('user_cardinality', other_fields)
                self.assertInlinedFields(form, [
                    ('seda_start_date', 'StartDateInlineEntityCreationFormView'),
                ])
                self.assertNoRemove(form, 'seda_start_date', 'object')
                self.assertNoTitle(form, 'seda_start_date', 'object')

    def test_create_data_object_full(self):
        with self.admin_access.web_request() as req:
            req.entity_from_eid(self.transfer_eid).cw_set(simplified_profile=False)
            req.cnx.commit()
            form = self.create_inlined_form(req, 'SEDADataObjectReference', formtype='inlined')
            self.assertEqual(entity_fields(form.fields),
                             ['seda_data_object_reference_id'])

    def test_create_data_object_simplified(self):
        with self.admin_access.web_request() as req:
            form = self.create_inlined_form(req, 'SEDADataObjectReference', formtype='inlined')
            self.assertEqual(entity_fields(form.fields),
                             [])

    def test_create_data_object_from_archive_unit_component(self):
        with self.admin_access.web_request() as req:
            unit = testutils.create_archive_unit(None, cnx=req.cnx)[0]
            req.cnx.commit()
            form = self.create_inlined_form(req, 'SEDADataObjectReference', formtype='inlined',
                                            parent=unit.eid)
            self.assertEqual(entity_fields(form.fields),
                             [])


class CloneActionsTC(CubicWebTC):

    def test_archive_transfer(self):
        with self.admin_access.web_request() as req:
            transfer = req.create_entity('SEDAArchiveTransfer', title=u'test')
            unit = testutils.create_archive_unit(transfer)[0]
            req.cnx.commit()

            actions = self.pactionsdict(req, transfer.as_rset())
            self.assertIn(clone.SEDAArchiveTransferCloneAction, actions['mainactions'])

            actions = self.pactionsdict(req, unit.as_rset())
            self.assertNotIn(clone.SEDAArchiveUnitCloneAction, actions['mainactions'])

    def test_unit_component(self):
        with self.admin_access.web_request() as req:
            unit, unit_alt, unit_alt_seq = testutils.create_archive_unit(None, cnx=req)
            subunit = testutils.create_archive_unit(unit_alt_seq)[0]
            req.cnx.commit()

            actions = self.pactionsdict(req, unit.as_rset())
            self.assertIn(clone.SEDAArchiveUnitCloneAction, actions['mainactions'])

            actions = self.pactionsdict(req, subunit.as_rset())
            self.assertNotIn(clone.SEDAArchiveUnitCloneAction, actions['mainactions'])


class CloneImportTC(CubicWebTC):
    """Tests for 'seda.doimport' controller (called from JavaScript)."""

    def setup_database(self):
        with self.admin_access.cnx() as cnx:
            transfer = cnx.create_entity('SEDAArchiveTransfer', title=u'test')
            unit, unit_alt, unit_alt_seq = testutils.create_archive_unit(None, cnx=cnx,
                                                                         user_cardinality=u'1',
                                                                         user_annotation=u'plop')
            bdo = testutils.create_data_object(unit_alt_seq, filename=u'file.txt')
            self.transfer_eid = transfer.eid
            self.unit_eid = unit.eid
            self.bdo_eid = bdo.eid
            cnx.commit()

    def doctypecode(self, cnx):
        return cnx.create_entity(
            'SEDADocumentTypeCode',
            seda_document_type_code_value=self.doctypecodevalue_eid)

    def test_import_one_entity(self):
        params = dict(eid=text_type(self.transfer_eid),
                      cloned=text_type(self.unit_eid))
        with self.admin_access.web_request(**params) as req:
            path, _ = self.expect_redirect_handle_request(
                req, 'seda.doimport')
            etype, eid = path.split('/')
            self.assertEqual(etype, 'SEDAArchiveTransfer'.lower())
            clone = req.execute('Any X WHERE X seda_archive_unit P, P eid %(p)s',
                                {'p': eid}).one()
            self.assertEqual([x.eid for x in clone.clone_of], [self.unit_eid])
            # Check that original entity attributes have been copied.
            self.assertEqual(clone.user_cardinality, u'1')
            self.assertEqual(clone.user_annotation, u'plop')
            # Check that data object has been copied and linked to the transfer
            seq = clone.first_level_choice.content_sequence
            bdo = seq.reverse_seda_data_object_reference[0].seda_data_object_reference_id[0]
            self.assertNotEqual(bdo.eid, self.bdo_eid)
            self.assertEqual(bdo.filename, 'file.txt')
            transfer = req.entity_from_eid(self.transfer_eid)
            transfer_bdos = [do.eid for do in transfer.reverse_seda_binary_data_object]
            self.assertEqual(transfer_bdos, [bdo.eid])

    def test_import_multiple_entities(self):
        with self.admin_access.cnx() as cnx:
            unit, unit_alt, unit_alt_seq = testutils.create_archive_unit(
                None, cnx=cnx,
                user_cardinality=u'0..1', user_annotation=u'plouf')
            cnx.commit()
        to_clone = [self.unit_eid, unit.eid]
        params = dict(eid=text_type(self.transfer_eid),
                      cloned=','.join([text_type(self.unit_eid), text_type(unit.eid)]))
        with self.admin_access.web_request(**params) as req:
            path, _ = self.expect_redirect_handle_request(
                req, 'seda.doimport')
            etype, eid = path.split('/')
            self.assertEqual(etype, 'SEDAArchiveTransfer'.lower())
            rset = req.execute(
                'Any X,O WHERE X seda_archive_unit P, P eid %(p)s, X clone_of O',
                {'p': eid})
            self.assertEqual(len(rset), 2)
            self.assertCountEqual([oeid for __, oeid in rset.rows], to_clone)
            cardinalities, annotations = zip(*[
                (clone.user_cardinality, clone.user_annotation)
                for clone in rset.entities()])
            self.assertCountEqual(cardinalities, ('1', '0..1'))
            self.assertCountEqual(annotations, ('plop', 'plouf'))


class SEDATreeTC(CubicWebTC):

    def setup_database(self):
        with self.admin_access.cnx() as cnx:
            self.transfer_eid = cnx.create_entity('SEDAArchiveTransfer',
                                                  title=u'Test',
                                                  simplified_profile=True).eid
            cnx.commit()

    def test_archiveunit_reparent_to_transfer(self):
        with self.admin_access.cnx() as cnx:
            transfer = cnx.entity_from_eid(self.transfer_eid)
            unit, unit_alt, unit_alt_seq = testutils.create_archive_unit(transfer)
            unit2 = testutils.create_archive_unit(unit_alt_seq)[0]
            cnx.commit()
            unit2.cw_adapt_to('IJQTree').reparent(transfer.eid, 0)
            cnx.commit()
            transfer.cw_clear_all_caches()
            self.assertEqual([x.eid for x in transfer.reverse_seda_archive_unit],
                             [unit2.eid, unit.eid])

    def test_archiveunit_reparent_to_archiveunit(self):
        with self.admin_access.cnx() as cnx:
            transfer = cnx.entity_from_eid(self.transfer_eid)
            unit, _, unit_alt_seq = testutils.create_archive_unit(transfer)
            unit2, _, unit2_alt_seq = testutils.create_archive_unit(transfer)
            subunit, _, _ = testutils.create_archive_unit(unit2_alt_seq)
            cnx.commit()
            unit.cw_clear_all_caches()
            unit.cw_adapt_to('IJQTree').reparent(unit2.eid, 0)
            cnx.commit()
            unit2_alt_seq.cw_clear_all_caches()
            self.assertEqual([x.eid for x in unit2_alt_seq.reverse_seda_archive_unit],
                             [unit.eid, subunit.eid])

    def test_binarydataobject_reparent(self):
        with self.admin_access.cnx() as cnx:
            transfer = cnx.entity_from_eid(self.transfer_eid)
            unit, _, unit_alt_seq = testutils.create_archive_unit(transfer)
            unit2, _, unit2_alt_seq = testutils.create_archive_unit(transfer)
            bdo = testutils.create_data_object(transfer)
            ref = cnx.create_entity('SEDADataObjectReference',
                                    seda_data_object_reference=unit_alt_seq,
                                    seda_data_object_reference_id=bdo)
            cnx.commit()
            bdo.cw_clear_all_caches()
            bdo.cw_adapt_to('IJQTree').reparent(unit2.eid, 0)
            cnx.commit()
            unit2_alt_seq.cw_clear_all_caches()
            self.assertEqual([x.eid for x in unit2_alt_seq.reverse_seda_data_object_reference],
                             [ref.eid])
            unit_alt_seq.cw_clear_all_caches()
            self.assertFalse(unit_alt_seq.reverse_seda_data_object_reference)

    def test_binarydataobject_reorder_full(self):
        with self.admin_access.cnx() as cnx:
            transfer = cnx.entity_from_eid(self.transfer_eid)
            transfer.cw_set(simplified_profile=False)
            bdo1 = testutils.create_data_object(transfer)
            bdo2 = testutils.create_data_object(transfer)
            cnx.commit()

            transfer.cw_adapt_to('IJQTree').move_child_at_index(bdo2.eid, 0)

            transfer.cw_clear_all_caches()
            self.assertEqual([x.eid for x in transfer.reverse_seda_binary_data_object],
                             [bdo2.eid, bdo1.eid])

            bdo2.cw_clear_all_caches()
            transfer.cw_adapt_to('IJQTree').move_child_at_index(bdo2.eid, 2)

            bdo1.cw_clear_all_caches()
            bdo2.cw_clear_all_caches()

            transfer.cw_clear_all_caches()
            self.assertEqual([x.eid for x in transfer.cw_adapt_to('ITreeBase').iterchildren()],
                             [bdo1.eid, bdo2.eid])

    def test_binarydataobject_reorder_simplified(self):
        with self.admin_access.cnx() as cnx:
            transfer = cnx.entity_from_eid(self.transfer_eid)
            unit, _, unit_alt_seq = testutils.create_archive_unit(transfer)
            bdo1 = testutils.create_data_object(unit_alt_seq)
            bdo2 = testutils.create_data_object(unit_alt_seq)
            cnx.commit()

            unit.cw_adapt_to('IJQTree').move_child_at_index(bdo2.eid, 0)

            transfer.cw_clear_all_caches()
            self.assertEqual([x.eid for x in unit.cw_adapt_to('ITreeBase').iterchildren()],
                             [bdo2.eid, bdo1.eid])


class FakeForm(object):
    def __init__(self, req, edited_entity):
        self._cw = req
        self.edited_entity = edited_entity


class ArchiveUnitVocabularyTC(CubicWebTC):

    def test_unit_ref_vocabulary(self):
        with self.admin_access.web_request() as req:
            transfer = req.cnx.create_entity('SEDAArchiveTransfer', title=u'Test')
            archunit, _, alt_seq = testutils.create_archive_unit(transfer)
            req.cnx.commit()
            archunit.cw_clear_all_caches()
            # actually expect a archive unit reference, but we want to test the query so any object
            # linked to the container is fine
            form = FakeForm(req, archunit)
            self.assertEqual(archiveunit.unit_ref_vocabulary(form, None),
                             [(u'archive unit title', text_type(archunit.eid))])

    def test_do_ref_vocabulary(self):
        with self.admin_access.web_request() as req:
            transfer = req.cnx.create_entity('SEDAArchiveTransfer', title=u'Test')
            bdo = testutils.create_data_object(transfer)
            req.cnx.commit()
            bdo.cw_clear_all_caches()
            # actually expect a data object reference, but we want to test the query so any object
            # linked to the container is fine
            form = FakeForm(req, bdo)
            self.assertEqual(archiveunit.do_ref_vocabulary(form, None),
                             [(u'data object title', text_type(bdo.eid))])


class SEDANavigationTC(CubicWebTC):

    def test_breadcrumbs(self):
        with self.admin_access.cnx() as cnx:
            transfer = cnx.create_entity('SEDAArchiveTransfer', title=u'test profile')
            unit1 = testutils.create_archive_unit(transfer)[0]
            unit2 = testutils.create_archive_unit(None, cnx=cnx)[0]
            cnx.commit()
        with self.admin_access.web_request() as req:
            unit1 = req.entity_from_eid(unit1.eid)
            unit2 = req.entity_from_eid(unit2.eid)
            # unit1 is related to a transfer
            breadcrumbs = unit1.cw_adapt_to('IBreadCrumbs').breadcrumbs()
            expected_breadcrumbs = [transfer, unit1]
            self.assertEqual(breadcrumbs, expected_breadcrumbs)
            # unit2 is not related to a transfer, breadcrumbs leads to /sedalib.
            breadcrumbs = unit2.cw_adapt_to('IBreadCrumbs').breadcrumbs()
            expected_breadcrumbs = [
                (u'http://testing.fr/cubicweb/sedalib', u'SEDAArchiveUnit_plural'),
                unit2,
            ]
            self.assertEqual(breadcrumbs, expected_breadcrumbs)


if __name__ == '__main__':
    unittest.main()



# copyright 2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-seda versions compatibility diagnosis test."""

from cubicweb.devtools.testlib import CubicWebTC

from cubicweb_seda import testutils


class CompatAnalyzerTC(CubicWebTC):

    def test_rules(self):
        with self.admin_access.cnx() as cnx:
            create = cnx.create_entity

            transfer = create('SEDAArchiveTransfer', title=u'diagnosis testing')
            doctor = transfer.cw_adapt_to('ISEDACompatAnalyzer')
            cnx.commit()
            self.assertFailingRules(doctor, ['SEDA 2.0', 'SEDA 1.0', 'SEDA 0.2', 'simplified'])

            unit, unit_alt, unit_alt_seq = testutils.create_archive_unit(transfer)
            transfer.cw_clear_all_caches()
            cnx.commit()
            self.assertFailingRules(doctor, ['SEDA 2.0', 'SEDA 0.2', 'simplified'],
                                    'seda1_need_access_rule')

            access_rule = create('SEDAAccessRule', seda_access_rule=unit_alt_seq)
            unit_alt_seq.cw_clear_all_caches()
            cnx.commit()
            self.assertFailingRules(doctor, ['SEDA 2.0'], 'rule_without_rule')

            access_rule_seq = create('SEDASeqAccessRuleRule',
                                     user_cardinality=u'1..n',
                                     reverse_seda_seq_access_rule_rule=access_rule)
            access_rule.cw_clear_all_caches()
            cnx.commit()
            self.assertFailingRules(doctor, ['SEDA 2.0'],
                                    'rule_unsupported_card', 'rule_need_start_date')

            access_rule_seq.cw_set(user_cardinality=u'1')
            start_date = create('SEDAStartDate',
                                user_cardinality=u'0..1',
                                seda_start_date=access_rule_seq)
            access_rule_seq.cw_clear_all_caches()
            cnx.commit()
            self.assertFailingRules(doctor, ['SEDA 2.0'],
                                    'rule_start_unsupported_card')

            start_date.cw_set(user_cardinality=u'1')
            start_date.cw_clear_all_caches()
            access_rule_seq2 = create('SEDASeqAccessRuleRule',
                                      user_cardinality=u'1..n',
                                      reverse_seda_seq_access_rule_rule=access_rule)
            access_rule.cw_clear_all_caches()
            cnx.commit()
            self.assertFailingRules(doctor, ['SEDA 2.0'],
                                    'rule_with_too_much_rules')

            access_rule_seq2.cw_delete()
            inherit_ctl = create('SEDAAltAccessRulePreventInheritance',
                                 reverse_seda_alt_access_rule_prevent_inheritance=access_rule)
            create('SEDARefNonRuleId', seda_ref_non_rule_id_from=inherit_ctl)
            access_rule.cw_clear_all_caches()
            cnx.commit()
            self.assertFailingRules(doctor, ['SEDA 2.0'],
                                    'rule_ref_non_rule_id')

    def test_custodial_history(self):
        with self.admin_access.cnx() as cnx:
            create = cnx.create_entity

            transfer = create('SEDAArchiveTransfer', title=u'diagnosis testing')
            unit, unit_alt, unit_alt_seq = testutils.create_archive_unit(transfer)
            access_rule = create('SEDAAccessRule',
                                 seda_access_rule=unit_alt_seq)
            access_rule_seq = create('SEDASeqAccessRuleRule',
                                     reverse_seda_seq_access_rule_rule=access_rule)
            create('SEDAStartDate',
                   user_cardinality=u'1',
                   seda_start_date=access_rule_seq)

            history_item = create('SEDACustodialHistoryItem',
                                  seda_custodial_history_item=unit_alt_seq)

            doctor = transfer.cw_adapt_to('ISEDACompatAnalyzer')
            cnx.commit()
            self.assertFailingRules(doctor, ['SEDA 2.0', 'SEDA 1.0', 'SEDA 0.2', 'simplified'])

            history_item2 = create('SEDACustodialHistoryItem',
                                   seda_custodial_history_item=unit_alt_seq)
            unit_alt_seq.cw_clear_all_caches()
            cnx.commit()
            self.assertFailingRules(doctor, ['SEDA 2.0', 'SEDA 1.0', 'simplified'],
                                    'seda02_custodial_history_items')

            history_item2.cw_delete()
            create('SEDAwhen', seda_when=history_item)
            unit_alt_seq.cw_clear_all_caches()
            history_item.cw_clear_all_caches()
            cnx.commit()
            self.assertFailingRules(doctor, ['SEDA 2.0', 'SEDA 1.0', 'simplified'],
                                    'seda02_custodial_history_when')

    def test_archive_unit_reference_in_transfer(self):
        with self.admin_access.cnx() as cnx:
            transfer = cnx.create_entity('SEDAArchiveTransfer', title=u'diagnosis testing')
            testutils.create_archive_unit(transfer, archive_unit_reference=True)

            doctor = transfer.cw_adapt_to('ISEDACompatAnalyzer')
            cnx.commit()
            self.assertFailingRules(doctor, ['SEDA 2.0'], 'use_archive_unit_ref')

    def test_archive_unit_reference_in_archive_unit(self):
        with self.admin_access.cnx() as cnx:
            transfer = cnx.create_entity('SEDAArchiveTransfer', title=u'diagnosis testing')
            unit, unit_alt, unit_alt_seq = testutils.create_archive_unit(transfer)
            testutils.create_archive_unit(unit_alt_seq, archive_unit_reference=True)

            doctor = transfer.cw_adapt_to('ISEDACompatAnalyzer')
            cnx.commit()
            self.assertFailingRules(doctor, ['SEDA 2.0'],
                                    'use_archive_unit_ref', 'seda1_need_access_rule')

    def assertFailingRules(self, doctor, expected_formats, *expected_rule_ids):
        doctor.entity.cw_clear_all_caches()
        rule_ids = set(rule_id for rule_id, entity in doctor.failing_rules())
        self.assertEqual(rule_ids, set(expected_rule_ids))
        expected_formats.append('RNG')
        self.assertEqual(doctor.diagnose(), set(expected_formats))
        self.assertEqual(doctor.entity.compat_list, ', '.join(sorted(expected_formats)))

    # in tests below, assertIsRNGAmbiguous is used to test that hooks have
    # properly detected the error (or fix) while assertRNGAmbiguousDiagnostic
    # will test the underlying details of the diagnosis tool.

    def test_multiple_child_unhandled_cardinality_archive_unit(self):
        with self.admin_access.cnx() as cnx:
            transfer = cnx.create_entity('SEDAArchiveTransfer', title=u'test profile')
            doctor = transfer.cw_adapt_to('ISEDACompatAnalyzer')
            testutils.create_archive_unit(transfer)
            unit2 = testutils.create_archive_unit(transfer, user_cardinality=u'0..1')[0]
            testutils.create_archive_unit(transfer, user_cardinality=u'1..n')
            cnx.commit()

            self.assertIsRNGAmbiguous(transfer, True)
            self.assertErrorOn(doctor, 'rng_ambiguity', [(transfer, 'seda_archive_units_tab')])

            unit2.cw_set(user_cardinality=u'1')
            cnx.commit()
            self.assertIsRNGAmbiguous(transfer, False)

            unit3 = testutils.create_archive_unit(transfer, user_cardinality=u'0..1')[0]
            cnx.commit()
            self.assertIsRNGAmbiguous(transfer, True)
            self.assertErrorOn(doctor, 'rng_ambiguity', [(transfer, 'seda_archive_units_tab')])

            unit3.cw_delete()
            cnx.commit()
            self.assertIsRNGAmbiguous(transfer, False)

    def test_multiple_child_unhandled_cardinality_sub_archive_unit(self):
        with self.admin_access.cnx() as cnx:
            transfer = cnx.create_entity('SEDAArchiveTransfer', title=u'test profile')
            doctor = transfer.cw_adapt_to('ISEDACompatAnalyzer')
            unit, unit_alt, unit_alt_seq = testutils.create_archive_unit(transfer)
            testutils.create_archive_unit(unit_alt_seq, user_cardinality=u'1..n')
            testutils.create_archive_unit(unit_alt_seq, user_cardinality=u'1..n')
            cnx.commit()

            self.assertIsRNGAmbiguous(transfer, True)
            self.assertErrorOn(doctor, 'rng_ambiguity', [(unit, 'seda_archive_units_tab')])

    def test_multiple_child_unhandled_cardinality_document(self):
        with self.admin_access.cnx() as cnx:
            transfer = cnx.create_entity('SEDAArchiveTransfer', title=u'test profile')
            doctor = transfer.cw_adapt_to('ISEDACompatAnalyzer')
            unit, unit_alt, unit_alt_seq = testutils.create_archive_unit(transfer)
            testutils.create_data_object(unit_alt_seq, user_cardinality=u'0..n',
                                         seda_binary_data_object=transfer)
            bdo = testutils.create_data_object(unit_alt_seq, user_cardinality=u'1..n',
                                               seda_binary_data_object=transfer)
            cnx.commit()
            self.assertIsRNGAmbiguous(transfer, True)
            self.assertErrorOn(doctor, 'rng_ambiguity', [(unit, 'seda_data_objects_tab'),
                                                         (transfer, 'seda_data_objects_tab')])
            _delete_data_object(bdo)
            cnx.commit()
            self.assertIsRNGAmbiguous(transfer, False)

            # test data objects cardinality is considered globally to the profil
            # **if it's not a simplified profile**
            unit2, unit2_alt, unit2_alt_seq = testutils.create_archive_unit(transfer)
            bdo = testutils.create_data_object(unit2_alt_seq, user_cardinality=u'0..n',
                                               seda_binary_data_object=transfer)
            cnx.commit()
            self.assertIsRNGAmbiguous(transfer, True)
            self.assertErrorOn(doctor, 'rng_ambiguity', [(transfer, 'seda_data_objects_tab')])
            _delete_data_object(bdo)
            cnx.commit()
            self.assertIsRNGAmbiguous(transfer, False)

            # though adding a data object with card != 1 to another archive unit
            # **on a simplified profile** is fine
            transfer.cw_set(simplified_profile=True)
            testutils.create_data_object(unit2_alt_seq, user_cardinality=u'0..n',
                                         seda_binary_data_object=transfer)
            cnx.commit()
            self.assertErrorOn(doctor, 'rng_ambiguity', [])
            self.assertIsRNGAmbiguous(transfer, False)

    def test_multiple_child_unhandled_cardinality_keyword(self):
        with self.admin_access.cnx() as cnx:
            transfer = cnx.create_entity('SEDAArchiveTransfer', title=u'test profile')
            doctor = transfer.cw_adapt_to('ISEDACompatAnalyzer')
            unit, unit_alt, unit_alt_seq = testutils.create_archive_unit(transfer)
            cnx.create_entity('SEDAKeyword', seda_keyword=unit_alt_seq,
                              user_cardinality=u'1..n')
            cnx.create_entity('SEDAKeyword', seda_keyword=unit_alt_seq,
                              user_cardinality=u'1..n')
            cnx.commit()

            self.assertErrorOn(doctor, 'rng_ambiguity', [(unit, 'seda_indexation_tab')])

    def test_single_cardinality_not_first(self):
        with self.admin_access.cnx() as cnx:
            transfer = cnx.create_entity('SEDAArchiveTransfer', title=u'test profile')
            doctor = transfer.cw_adapt_to('ISEDACompatAnalyzer')
            u1 = testutils.create_archive_unit(transfer, user_cardinality=u'1..n')
            u2 = testutils.create_archive_unit(transfer, user_cardinality=u'1')
            cnx.commit()
            self.assertEqual(u1[0].ordering, 1)
            self.assertEqual(u2[0].ordering, 2)
            self.assertEqual(u1[0].user_cardinality, '1..n')
            self.assertEqual(u2[0].user_cardinality, '1')
            self.assertIsRNGAmbiguous(transfer, True)
            self.assertErrorOn(doctor, 'rng_mandatory_not_first',
                               [(transfer, 'seda_archive_units_tab')])

    def assertErrorOn(self, doctor, error_id, expected_errors):
        errors = set()
        for error in doctor.detect_problems():
            if error.rule_id == error_id:
                errors.add((error.entity, error.tab_id))

        self.assertEqual(errors, set(expected_errors))

    def assertIsRNGAmbiguous(self, transfer, ambiguous):
        transfer.cw_clear_all_caches()
        if ambiguous:
            self.assertNotIn('RNG', transfer.compat_list)
        else:
            self.assertIn('RNG', transfer.compat_list)


def _delete_data_object(data_object):
    for ref in data_object.reverse_seda_data_object_reference_id:
        ref.cw_delete()
    data_object.cw_delete()


if __name__ == '__main__':
    import unittest
    unittest.main()



# copyright 2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-seda tools to diagnose versions compatibility of profiles."""

from collections import namedtuple
from itertools import chain

from cubicweb import _
from cubicweb.view import EntityAdapter
from cubicweb.predicates import is_instance


ALL_FORMATS = frozenset(('SEDA 2.0', 'SEDA 1.0', 'SEDA 0.2', 'simplified', 'RNG'))

Rule = namedtuple('Rule', ['impacted_formats', 'message', 'tab_id', 'watch'])
RULES = {
    'use_archive_unit_ref': Rule(
        set(['SEDA 1.0', 'SEDA 0.2', 'simplified']),
        _("Archive unit reference are only supported by SEDA 2."),
        'main_tab',
        set([
            'seda_seq_access_rule_rule',
            'seda_seq_appraisal_rule_rule',
        ])),

    'seda1_need_access_rule': Rule(
        set(['SEDA 1.0']),
        _("First level archive unit must have an associated access rule to be exportable "
          "in SEDA 1. You may define it on the archive unit or as a default rule on the "
          "transfer."),
        'seda_management_tab',
        set([
            'seda_archive_unit',
            'seda_alt_archive_unit_archive_unit_ref_id',
            'seda_seq_alt_archive_unit_archive_unit_ref_id_management',
            'seda_access_rule',
        ])),
    'rule_without_rule': Rule(
        set(['SEDA 1.0', 'SEDA 0.2', 'simplified']),
        _("Some management rule has no inner rule, one is required."),
        'seda_management_tab',
        set([
            'seda_seq_access_rule_rule',
            'seda_seq_appraisal_rule_rule',
        ])),
    'rule_with_too_much_rules': Rule(
        set(['SEDA 1.0', 'SEDA 0.2', 'simplified']),
        _("Some management rule has more than one inner rules, a single one is required."),
        'seda_management_tab',
        set([
            'seda_seq_access_rule_rule',
            'seda_seq_appraisal_rule_rule',
        ])),
    'rule_unsupported_card': Rule(
        set(['SEDA 1.0', 'SEDA 0.2', 'simplified']),
        _("Inner rule has cardinality other than 1."),
        'seda_management_tab',
        set([
            ('SEDASeqAccessRuleRule', 'user_cardinality'),
            ('SEDASeqAppraisalRuleRule', 'user_cardinality'),
        ])),
    'rule_need_start_date': Rule(
        set(['SEDA 1.0', 'SEDA 0.2', 'simplified']),
        _("Inner rule has no start date."),
        'seda_management_tab',
        set([
            'seda_start_date',
        ])),
    'rule_start_unsupported_card': Rule(
        set(['SEDA 1.0', 'SEDA 0.2', 'simplified']),
        _("Start date has cardinality other than 1."),
        'seda_management_tab',
        set([
            ('SEDAStartDate', 'user_cardinality'),
        ])),
    'rule_ref_non_rule_id': Rule(
        set(['SEDA 1.0', 'SEDA 0.2', 'simplified']),
        _("Rule has an explicit rule deactivation - only ignore all rules is supported in "
          "simplified profiles."),
        'seda_management_tab',
        set([
            'seda_ref_non_rule_id_from',
        ])),

    'seda02_custodial_history_items': Rule(
        set(['SEDA 0.2']),
        _("Custodial history is text with SEDA 0.2, hence only one item element is considered."),
        'seda_history_tab',
        set([
            'seda_custodial_history_item',
        ])),
    'seda02_custodial_history_when': Rule(
        set(['SEDA 0.2']),
        _("Custodial history is text with SEDA 0.2, hence date information isn't considered."),
        'seda_history_tab',
        set([
            'seda_when',
        ])),
    'rng_ambiguity': Rule(
        set(['RNG']),
        _("More than one children with cardinality not equal to '1', "
          "this may cause RNG validation problems."),
        None,  # generated at run time
        []),  # have it's own hook
    'rng_mandatory_not_first': Rule(
        set(['RNG']),
        _("Children with cardinality equal to '1' should appear before other "
          "to avoid potential RNG validation problems."),
        None,  # generated at run time
        []),  # have it's own hook
}


class CompatError(namedtuple('_CompatError', [
        'impacted_formats', 'message', 'tab_id', 'entity', 'rule_id'])):
    """Convenience class holding information about a problem in a profile forbidding export to some
    format:

    * `impacted_formats`: set of formats that are no more available because of this error,
    * `message`: message string explaining the problem,
    * `entity`: 1st class entity where the error lies (one of archive unit, data object, etc.),
    * `tab`: entity's tab where the problem may be fixed.
    """
    def __new__(cls, rule_id, entity, **kwargs):
        rule = RULES[rule_id]
        impacted_formats = kwargs.get('impacted_formats', rule.impacted_formats)
        message = kwargs.get('message', rule.message)
        tab_id = kwargs.get('tab_id', rule.tab_id)
        return super(CompatError, cls).__new__(
            cls, impacted_formats, message, tab_id, entity, rule_id)


class ISEDACompatAnalyzer(EntityAdapter):
    """Adapter that will analyze profile to diagnose its format compatibility (SEDA 2, SEDA 1 and/or
    SEDA 0.2) while being to explain the problem to end-users.
    """

    __regid__ = 'ISEDACompatAnalyzer'
    __select__ = is_instance('SEDAArchiveTransfer')

    def diagnose(self):
        possible_formats = set(ALL_FORMATS)
        for compat_error in self.detect_problems():
            possible_formats -= compat_error.impacted_formats
        return possible_formats

    def detect_problems(self):
        """Yield :class:`CompatError` describing a problem that prevents the given profile to be
        compatible with some format.
        """
        for args in self.failing_rules():
            assert len(args) in (2, 3)
            extra_kwargs = args[2] if len(args) == 3 else {}
            yield CompatError(*args[:2], **extra_kwargs)

    def failing_rules(self):
        """Yield (rule identifier, problematic entity) describing a problem that prevents the given
        profile to be compatible with some format.
        """
        for check_method in (self._check_usage_of_reference_archive_unit,
                             self._check_management_rules,
                             self._check_custodial_history_rules,
                             self._check_rng_ambiguities):
            for problem in check_method():
                yield problem

    def _check_management_rules(self):  # noqa (too complex)
        profile = self.entity
        # First level archive unit needs an access rule (SEDA 1)
        if not profile.reverse_seda_access_rule:
            for archive_unit in profile.archive_units:
                seq = archive_unit.first_level_choice.content_sequence
                if seq is None:
                    # reference archive unit
                    continue
                if not seq.reverse_seda_access_rule:
                    yield 'seda1_need_access_rule', archive_unit
        # Access/appraisal rule must have one and only one sequence, which have one and only one
        # start date (and both of them must have 1 cardinality)
        for rule in self._cw.execute(
                'Any X WHERE X is IN (SEDAAccessRule, SEDAAppraisalRule), '
                'X container C, C eid %(c)s', {'c': profile.eid}).entities():
            if not rule.rules:
                yield 'rule_without_rule', _parent(rule)
            elif len(rule.rules) > 1:
                yield 'rule_with_too_much_rules', _parent(rule)
            else:
                rule_seq = rule.rules[0]
                if rule_seq.user_cardinality != '1':
                    yield 'rule_unsupported_card', _parent(rule)
                if not rule_seq.start_date:
                    yield 'rule_need_start_date', _parent(rule)
                elif rule_seq.start_date.user_cardinality != '1':
                    yield 'rule_start_unsupported_card', _parent(rule)
        # Access/appraisal rule shouldn't use ref non rule id
        for rule_type in ('access', 'appraisal'):
            for rule in self._cw.execute(
                    'DISTINCT Any X WHERE X seda_alt_{0}_rule_prevent_inheritance ALT,'
                    'REF seda_ref_non_rule_id_from ALT, '
                    'X container C, C eid %(c)s'.format(rule_type),
                    {'c': profile.eid}).entities():
                yield 'rule_ref_non_rule_id', _parent(rule)

    def _check_custodial_history_rules(self):
        profile = self.entity
        # Check for more than one custodial history items (SEDA 0.2)
        for content in self._cw.execute(
                'Any CONT GROUPBY CONT WHERE X seda_custodial_history_item CONT,'
                'X container C, C eid %(c)s HAVING COUNT(X) > 1',
                {'c': profile.eid}).entities():
            parent = _parent(content) if profile.simplified_profile else content
            yield 'seda02_custodial_history_items', parent
        # Check for SEDAwhen usage (SEDA 0.2)
        for content in self._cw.execute(
                'DISTINCT Any CONT WHERE X seda_custodial_history_item CONT,'
                'W seda_when X, X container C, C eid %(c)s',
                {'c': profile.eid}).entities():
            parent = _parent(content) if profile.simplified_profile else content
            yield 'seda02_custodial_history_when', parent

    def _check_usage_of_reference_archive_unit(self):
        for archive_unit in self._cw.execute(
                'DISTINCT Any X WHERE X is SEDAArchiveUnit,'
                'X seda_alt_archive_unit_archive_unit_ref_id ALT, '
                'NOT EXISTS(ALT seda_seq_alt_archive_unit_archive_unit_ref_id_management SEQ),'
                'X container C, C eid %(c)s', {'c': self.entity.eid}).entities():
            yield 'use_archive_unit_ref', archive_unit

    def _check_rng_ambiguities(self):
        for rql, error_id in [(_AMBIGUOUS_RQL, 'rng_ambiguity'),
                              (_MANDATORY_NOT_LAST_AMBIGUOUS_RQL, 'rng_mandatory_not_first')]:
            for eid, rtype in self._cw.execute(rql, {'c': self.entity.eid}):
                # if the container is a simplified profile, allow to have several
                # entities with cardinality != 1 under the seda_binary_data_object
                # relation because data objects are all linked to the transfer
                # through this relation, but this is only relevant for SEDA 2 export
                # and we don't want to prevent this in profiles for SEDA 0.2 and 1.0
                # export.
                if rtype == 'seda_binary_data_object' and self.entity.simplified_profile:
                    continue
                tab_id = _RTYPE_TO_TAB.get(rtype)
                yield error_id, _parent(self._cw.entity_from_eid(eid)), {'tab_id': tab_id}


def _parent(entity):
    """Return the first encountered parent which is an ArchiveUnit"""
    while entity.cw_etype not in ('SEDAArchiveTransfer', 'SEDAArchiveUnit',
                                  'SEDABinaryDataObject', 'SEDAPhysicalDataObject'):
        entity = entity.cw_adapt_to('IContained').parent
    return entity


_RTYPE_TO_TAB = {
    'seda_addressee_from': 'seda_agents_tab',
    'seda_archive_unit': 'seda_archive_units_tab',
    'seda_binary_data_object': 'seda_data_objects_tab',
    'seda_custodial_history_item': 'seda_history_tab',
    'seda_data_object_reference': 'seda_data_objects_tab',
    'seda_event': 'seda_events_tab',
    'seda_is_part_of': 'seda_relations_tab',
    'seda_is_version_of': 'seda_relations_tab',
    'seda_juridictional': 'seda_coverage_tab',
    'seda_keyword': 'seda_indexation_tab',
    'seda_physical_data_object': 'seda_data_objects_tab',
    'seda_recipient_from': 'seda_agents_tab',
    # don't care about order on this one, for now at least
    # 'seda_ref_non_rule_id_from': 'seda_management_tab',
    'seda_references': 'seda_relations_tab',
    'seda_related_transfer_reference': 'seda_at_related_transfers_tab',
    'seda_relationship': 'seda_do_relations',
    'seda_replaces': 'seda_relations_tab',
    'seda_requires': 'seda_relations_tab',
    'seda_spatial': 'seda_coverage_tab',
    'seda_tag': 'seda_indexation_tab',
    'seda_temporal': 'seda_coverage_tab',
    'seda_writer_from': 'seda_agents_tab',
}

_AMBIGUOUS_RQL = ' UNION '.join(chain(
    ('(Any P, "{rtype}" GROUPBY P WHERE X {rtype} P, P container C, C eid %(c)s, '
     'NOT X user_cardinality "1" HAVING COUNT(X) > 1)'.format(rtype=rtype)
     for rtype in _RTYPE_TO_TAB
     if rtype not in ('seda_binary_data_object',
                      'seda_physical_data_object',
                      'seda_related_transfer_reference')),
    # we need another RQL for top level entities without container relation
    ('(Any P, "{rtype}" GROUPBY P WHERE X {rtype} P, P eid %(c)s, '
     'NOT X user_cardinality "1" HAVING COUNT(X) > 1)'.format(rtype=rtype)
     for rtype in ('seda_archive_unit',
                   'seda_binary_data_object', 'seda_physical_data_object',
                   'seda_related_transfer_reference')),
))

_MANDATORY_NOT_LAST_AMBIGUOUS_RQL = ' UNION '.join(chain(
    ('(Any P, "{rtype}" WHERE EXISTS(X1 {rtype} P, X2 {rtype} P, '
     'P container C, C eid %(c)s, '
     'X1 user_cardinality "1", NOT X2 user_cardinality "1", '
     'X1 ordering ORD, X2 ordering < ORD))'.format(rtype=rtype)
     for rtype in _RTYPE_TO_TAB
     if rtype not in ('seda_binary_data_object',
                      'seda_physical_data_object',
                      'seda_related_transfer_reference')),
    # we need another RQL for top level entities without container relation
    ('(Any P, "{rtype}" WHERE EXISTS(X1 {rtype} P, X2 {rtype} P, P eid %(c)s, '
     'X1 user_cardinality "1", NOT X2 user_cardinality "1", '
     'X1 ordering ORD, X2 ordering < ORD))'.format(rtype=rtype)
     for rtype in ('seda_archive_unit',
                   'seda_binary_data_object', 'seda_physical_data_object',
                   'seda_related_transfer_reference')),
))



# copyright 2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-seda views for archive unit components"""

from cubicweb.predicates import (is_instance, multi_lines_rset,
                                 has_related_entities, relation_possible)
from cubicweb.web.views import basecomponents, baseviews, ibreadcrumbs, urlrewrite


class SEDAReqRewriter(urlrewrite.SimpleReqRewriter):
    ignore_baseclass_rules = True
    rules = [
        ('/sedalib', dict(rql='Any X WHERE X is SEDAArchiveUnit, '
                              'NOT X seda_archive_unit P',
                          vid='seda.sedalib')),
    ]


# navigation #######################################################################################

class SEDAComponentsBreadcrumbsAdapter(ibreadcrumbs.IBreadCrumbsAdapter):
    """Breadcrumbs adapter pointing to /sedalib route when entity has no
    parent.
    """
    __select__ = (ibreadcrumbs.IBreadCrumbsAdapter.__select__
                  & relation_possible('seda_archive_unit', role='subject')
                  & ~has_related_entities('seda_archive_unit', role='subject'))

    def breadcrumbs(self, *args, **kwargs):
        return [(self._cw.build_url('sedalib'), self._cw._('SEDAArchiveUnit_plural')),
                self.entity]


class SEDAComponentsBreadCrumbETypeVComponent(ibreadcrumbs.BreadCrumbEntityVComponent):
    """For proper display of the breadcrumb in the SEDA components list"""
    __select__ = (basecomponents.HeaderComponent.__select__
                  & multi_lines_rset() & is_instance('SEDAArchiveUnit'))

    def render_breadcrumbs(self, w, contextentity, path):
        w(u'<a href="%s">%s</a>' % (self._cw.build_url('sedalib'),
                                    self._cw._('SEDAArchiveUnit')))


# SEDA lib components ##############################################################################

class SEDALibView(baseviews.SameETypeListView):
    __regid__ = 'seda.sedalib'
    __select__ = is_instance('SEDAArchiveUnit')

    @property
    def title(self):
        return self._cw._('SEDAArchiveUnit_plural')
